;	
;	polarotor716.asm	
;	Satellite Polarotor servo driver for 16F716
;	PORTA:2 is servo pulse (.8 - 2.2 mS) out	
;	PORTA:0 is LNB volts and is a 8 bit A/D input
;	Math for volts display
;	Skew for each polarity is adjustable using rotary encoder
;	2x16 LCD display shows Polarity and Skew degrees
;	4 MHz xtal clock
;	V2 goal seek
;	Tweaked 15 mS delay, main loop timing
;	Defined constants for H&V thresholds
;	10/06/12
;	Bench functional
;*****************************************************************************
;
; Target Controller -      PIC16F716
;                          __________
;     PWM0_OUT--------RA2 |1       18| RA1---------
;            ---------RA3 |2       17| RA0---------LNBV_IN
;           ----------RA4 |3       16| RA7---------OSC
;     /MCLR-----------RA5 |4       15| RA6---------OSC
;     Ground----------Vss |5       14| VDD---------+5 V
;     ENCODER A-------RB0 |6       13| RB7---------LCD 14
;     ENCODER B-------RB1 |7       12| RB6---------LCD 13
;     LCD_e-----------RB2 |8       11| RB5---------LCD 12
;     LCD_rs----------RB3 |9       10| RB4---------LCD 11
;                          ----------
;	
; ****************************************************************************
; * Device type and options.                                                 *
; ****************************************************************************
;
        LIST    P=16F716
	ERRORLEVEL	-302
        include <P16F716.INC>
        __CONFIG        h'3fb1' 
;
;------------------------------------------------------------   
;	
	ORG	0
	goto	main
	ORG	4
	goto	interrupt_service
;
;------------------------------------------------------------
;
	cblock	0x20
	temp_W				;interrupt regs
	temp_S				;
;
        ren_new                   	;New status of encoder
        ren_old                   	;Old status of encoder
	volts_in			;AD result
        encod_out                  	;encoder lookup output
        enc_offset                  	;encoder lookup offset
;
	vert_skew			;servo pulse width
	horz_skew			;(0-90)
	goal_pw_L			;
	goal_pw_H			;
	xfer_pw_H			;
	xfer_pw_L			;
;
	pw_low				;
	pw_high				;pulse width	
;
	hex_hi				;hex-decimal
        hex_lo				;(30h) 
	ten_thousands			;	
	thousands			;
	hundreds			;
        tens                    	;
        ones                    	;
;
        ACCa_hi                 	;math regs
        ACCa_lo				;
        ACCb_hi				;
        ACCb_lo				; 
        math_count			;        
;
	lcd_data			;LCD display data                                
	lcd_count			;
;
	count1				;
	count2				;
	rotor				;alternator
	delay				;(40h)delay timer	
        FLAGS                   	;(41h)various flags
;
 	endc
;
;------------------------------------------------------------
;
#define	PWM0_OUT	PORTA,2			;PWM0 output
#define E_LINE          PORTB,2         	;LCD E
#define RS_LINE         PORTB,3         	;LCD RS
;
#define	POL_FLAG	FLAGS,0			;0=V  1=H
;
VTHRESH		equ	.154			;vert AD threshold
HTHRESH		equ	.133			;horz AD threshold
VFACT		equ	.110			;voltage display conversion			
SVO_V		equ	.1500			;servo +0 deg (1.500 mS)
SVO_H		equ	.800			;servo -90 deg (0.800 mS)	
PULSE_GAP	equ	.65535 - .22000		;gap between pulses
BEG_PULSE	equ	.1850			;starting PW
ADCFG		equ	b'10000001'		;AD Osc/32, AD:0 read, OFF
LINE_1		equ	0x80			;LCD addresses
LINE_2		equ	0xC0			;
;
;------------------------------------------------------------
;
;	W contains offset
hsign
	addwf	PCL,F
	dt	"HORZ",0				;
;
vsign
	addwf	PCL,F
	dt	"VERT",0				;
;
enc_dec
	addwf	PCL,F
	dt	.0,.255,.1,.0,.1,.0,.0,.255		;
	dt	.255,.0,.0,.1,.0,.1,.255,.0		;
;
;------------------------------------------------------------
;
main
	clrf	PORTA
	clrf	PORTB
        bsf     STATUS,RP0              ; switch to bank 1
        movlw   b'11101011'             ; port A I/O
        movwf   TRISA
        movlw   b'00000011'             ; port B I/O
        movwf   TRISB
        movlw   b'00000001'             ; WPU on, TMR0/4
        movwf   OPTION_REG		;
        movlw   b'00000001'             ; TMR1
        movwf   PIE1                    ; perif int setup
	movlw	b'00000100'
	movwf	ADCON1			;A:0,1,3 analog
;
        bcf     STATUS,RP0              ; back to page 0
;
        clrf    PIR1                    ; reset perif int
        movlw   b'00000000'             ; Ints off
	movwf	INTCON
	movlw	b'00000001'
	movwf	T1CON			;timer /1, on
	movlw	b'00000000'		;
	movwf	T2CON			;TMR2 post/1, off 
	movlw	ADCFG
	movwf	ADCON0			;AD Osc/32, AD:0 in, ON
	clrf	CCP1CON			;CCP off
	clrf	PWM1CON			;PWM off
	clrf	ECCPAS			;
;
        call    init_lcd          	;initialize the LCD
;
	clrf	rotor			;
	clrf 	FLAGS			;reset
;
	movlw	b'01000000'
	movwf	INTCON			;PIE1 enable (TMR1)
	movlw	.45
	movwf	vert_skew
	movwf	horz_skew		;preload to (+/- 45)
	movlw	high BEG_PULSE
	movwf	pw_high
	movlw	low BEG_PULSE
	movwf	pw_low
	call	deb_delay		;let things settle
;
;------------------------------------------------------------
;	
;	Reads A:0 input volts
;
start_here
	bsf	INTCON,GIE		;enable TMR1 interrupt
;
    	movfw   PORTB		        ;Read encoder state
    	andlw   0x03            	;mask unused bits
    	movwf   ren_old             	;store it for preload
;
;	Routine borrowed from Chuck McManis
;
read_encoder
	call	deb_delay		;delay
	call	deb_delay		;delay x2
	call	read_AD			;read coax voltage & do pol flag
;
	movlw	LINE_2 +.2
	call	liter_c			;line #2 +2
	movfw	volts_in		;display volts
	movwf	ACCa_lo			;
	movlw	VFACT			;
	movwf	ACCb_lo			;
	call	multiply_8		;do the math
	movfw	ACCb_lo
	movwf	hex_lo
	movfw	ACCb_hi
	movwf	hex_hi
	call	hd_convert		;convert to dec
	movfw	ten_thousands
	movwf	lcd_data
	call	liter_d			;send 10's
	movfw	thousands
	movwf	lcd_data
	call	liter_d			;send 1's
	movlw	"."
	movwf	lcd_data
	call	liter_d			;decimal pt
	movfw	hundreds
	movwf	lcd_data
	call	liter_d			;send .10ths
	movlw	"v"
	movwf	lcd_data
	call	liter_d			;send vdc
;
    	movfw   PORTB		        ;Read the encoder
    	andlw   0x03            	;mask unused bits
    	movwf   ren_new             	;store it
    	clrc			        ;carry = zero    
    	rlf     ren_old,F         	;rotate ren_old Left
    	rlf     ren_old,W         	;again 
    	iorwf   ren_new,W           	;inclusive OR in the current value
    	movwf   enc_offset        	;store lookup offset
;
    	movfw   ren_new           	;get last time
    	movwf   ren_old           	;and store it
;
	movfw	enc_offset		;offset to W
	andlw	0x0f			;mask upper nybble	
	call	enc_dec			;do lookup table
        movwf   encod_out          	;store returned value (255,0,1)
;
        movlw   LINE_1			;start of polarity display
        call    liter_c                 ;send command
;
	btfsc 	POL_FLAG		;test polarity flag
	goto	send_horz
;
;------------------------------------------------------------
;	Vertical Area
;
send_vert
	clrf	count1			;LCD data pointer
;
send_vert_loop	
	movfw	count1			;pointer
	call	vsign			;look up table for "VERT"
	movwf	lcd_data
	movfw	lcd_data		;test for zero
	bz	send_V_polarity		;done, exit
	call	liter_d			;send to LCD
	incf	count1,F		;inc pointer
	goto	send_vert_loop		;loop
;
send_V_polarity
	movfw	encod_out		;test add encoder update
	addwf	vert_skew,W		;test for < zero
	xorlw	.255			;hold at zero ("0")
	bz	showvskew               ;
	movfw	encod_out		;test add encoder update
	addwf	vert_skew,W		;test for 91
	xorlw	.91			;hold at "90"
	bz	showvskew		;
	movfw	encod_out		;add encoder update
	addwf	vert_skew,F
showvskew
        movlw   LINE_1 +.5		;start of skew
        call    liter_c                 ;
	movlw	"+"
	movwf	lcd_data
	call	liter_d			;send (+) to LCD	
	movfw	vert_skew		;send vert skew value (0-90)
	movwf	hex_lo
	call	hd_convert
	movfw	tens
	movwf	lcd_data
	call	liter_d			;send 10's
	movfw	ones
	movwf	lcd_data
	call	liter_d			;send 1's
;
calc_V_pw
	movfw	vert_skew		;xfer to mult regs
	movwf	goal_pw_L		;vert skew degrees x 8
	clrf	goal_pw_H		;high byte clear
	clrc				;carry clear
	rlf	goal_pw_L,F		;rotate left 
	rlf	goal_pw_H,F		;x2
	rlf	goal_pw_L,F		;
	rlf	goal_pw_H,F		;x4
	rlf	goal_pw_L,F		;
	rlf	goal_pw_H,F		;x8
;
	movlw	low SVO_V		;minimum vert pulse width
	addwf	goal_pw_L,F 		;add to goal
	skpnc				;
	incf	goal_pw_H,F		;if carry bit set, add one	
	movlw	high SVO_V		;1500 uS + skew
	addwf	goal_pw_H,F		;goal_pw_H/L
;
;	Goal is now contained in goal_pw_H/L			
;
	call	goal_seeker		;do goal seek
	goto	prep2send		;
;
;------------------------------------------------------------
;	Horizontal Area
;		
;
send_horz
	clrf	count1			;LCD data pointer
;
send_horz_loop
	movfw	count1
	call	hsign			;look up table for "HORZ"
	movwf	lcd_data
	movfw	lcd_data		;test for zero
	bz	send_H_polarity		;done, exit
	call	liter_d			;send to LCD
	incf	count1,F		;
	goto	send_horz_loop		;loop
;
send_H_polarity
        movlw   LINE_1 +.5		;start of skew
        call    liter_c                 ;
	movfw	encod_out		;test add encoder update
	addwf	horz_skew,W		;test for < zero
	xorlw	.255			;
	bz	showhskew               ;hold at zero ("0")
	movfw	encod_out		;test add encoder update
	addwf	horz_skew,W		;test for 91
	xorlw	.91			;hold at "90"
	bz	showhskew		;
	movfw	encod_out		;add encoder update
	addwf	horz_skew,F
showhskew
	movlw	"-"
	movwf	lcd_data		;displays as negative	
	call	liter_d			;send to LCD
	movfw	horz_skew
	movwf	hex_lo
	call	hd_convert
	movfw	tens
	movwf	lcd_data
	call	liter_d			;send 10's
	movfw	ones
	movwf	lcd_data
	call	liter_d			;send 1's
;
calc_H_pw
	movfw	horz_skew		;xfer to mult regs
	movwf	goal_pw_L		;horz skew degrees x 8
	clrf	goal_pw_H		;
	clrc
	rlf	goal_pw_L,F		;
	rlf	goal_pw_H,F		;x2
	rlf	goal_pw_L,F		;
	rlf	goal_pw_H,F		;x4
	rlf	goal_pw_L,F		;
	rlf	goal_pw_H,F		;x8
;
	movlw	low SVO_H		;minimum horz pulse width
	addwf	goal_pw_L,F 		;
	skpnc				;
	incf	goal_pw_H,F		;
	movlw	high SVO_H		;800 uS + skew
	addwf	goal_pw_H,F		;goal_pw_H/L
	call	goal_seeker		;
;
;	Disable interrupts here to avoid loading corrupt 
;	TMR1 pw data
;
prep2send
	bcf	INTCON,GIE
	btfsc	INTCON,GIE
	goto	prep2send		;insure ints off
;	

;
	movfw	pw_low			;transfer pw_high/low 
	movwf	xfer_pw_L		;to xfer_pw_H/L
	movfw	pw_high
	movwf	xfer_pw_H
;
        comf    xfer_pw_L,F		;invert results
        incf    xfer_pw_L,F		;allows for simple
        skpnz				;addition
        decf    xfer_pw_H,F
        comf    xfer_pw_H,F
	bsf	INTCON,GIE		;enable ints
	goto	read_encoder		;return to top
;
;
;============================================================
;						
;	Responsible for servo pulse width
;
interrupt_service
    	movwf   temp_W
	bcf	STATUS,RP0		;bank 0
    	swapf   STATUS,W
    	movwf   temp_S
	btfss	PIR1,TMR1IF		;insure TMR1 int
	goto	abort_int		;unknown int
T1_handler
	bcf	PIR1,TMR1IF		;clear flag
	btfss	rotor,0
	goto	no_pulse
	movfw	xfer_pw_L			;inv 
	addwf	TMR1L,F
	skpnc
	incf	TMR1H,F			;
	movfw	xfer_pw_H			;inv
	addwf	TMR1H,F
	bsf	PWM0_OUT		;pulse 0 high        
	goto	outta_here
;
no_pulse
	bcf	PWM0_OUT		;pulse goes low
	movlw	high PULSE_GAP
	movwf	TMR1H			;
	movlw	low PULSE_GAP
	movwf	TMR1L			;22 mS til next int
outta_here
	incf	rotor,F			;inc each cycle
abort_int
    	swapf   temp_S,W
    	movwf   STATUS
    	swapf   temp_W,F
    	swapf   temp_W,W
    	retfie                      	;return to program        
;
;------------------------------------------------------------
;               Subroutines
;------------------------------------------------------------
;	Assumes AD module already set up & enabled
;	Gets coax voltage and handles polarity flag
read_AD
	bsf	ADCON0,GO		;start PIC battery AD convert	
read_AD_loop
	nop				;wait 1 instruct
	btfsc	ADCON0,GO
	goto	read_AD_loop
;
	movfw	ADRES
	movwf	volts_in		;results
;
	sublw	VTHRESH			;high threshold (17v)
	bnc	hi_volts		;
	movlw	HTHRESH			;low threshold (14v)
	subwf	volts_in,W		;
	skpnc
	return				;indetermined - leave as is	
	bcf	POL_FLAG		;volts low, make vert
	return
hi_volts
	bsf	POL_FLAG		;volts hi, horz
	return
;
;------------------------------------------------------------
;	16 bit hex to decimal up to 65,535
;
hd_convert
        movlw   "0"                    	;all zero ascii
        movwf   ones
        movwf   tens
        movwf   hundreds
        movwf   thousands
	movwf	ten_thousands
;
tenthou_hd
	movlw	0x27
	subwf	hex_hi,W
	bnc	thou_hd
	movlw	0x27
	subwf	hex_hi,F
	movlw	0x10
	subwf	hex_lo,W
	bnc	tenthou_check
do_lo2
	movlw	0x10
	subwf	hex_lo,F
	incf	ten_thousands,F		;10,000's +1
	goto	tenthou_hd
tenthou_check
	movfw	hex_hi
	skpnz
	goto	add_back_10k
        decf    hex_hi,F
        goto    do_lo2
add_back_10k
	movlw	0x27
	addwf	hex_hi,F
;
thou_hd
        movlw   0x03
        subwf   hex_hi,W
        bnc     hund_hd
        movlw   0x03
        subwf   hex_hi,F
        movlw   0xe8
        subwf   hex_lo,W
        bnc     thou_check
do_lo1
        movlw   0xe8
        subwf   hex_lo,F
        incf    thousands,F             ;1000's + 1
        goto    thou_hd
thou_check
        movfw   hex_hi
        skpnz
        goto    add_back_1k
        decf    hex_hi,F
        goto    do_lo1
add_back_1k
        movlw   0x03
        addwf   hex_hi,F
;
hund_hd
        movlw   0x64
        subwf   hex_lo,W		;hex_hi < 100?
        bc      do_hund			;no
        movfw   hex_hi			;yes, can we carry?
        skpnz
        goto    tens_hd			;no
        decf    hex_hi,F		;yes
do_hund
        movlw   0x64
        subwf   hex_lo,F
        incf    hundreds,F             	;100's + 1
        goto    hund_hd
;
;	count how many times 10 can be subtracted from hex_lo
;
tens_hd
        movlw   0x0a
        subwf   hex_lo,W
        bnc     do_ones
        movlw   0x0a
        subwf   hex_lo,F	
        incf    tens,F                 	;10's + 1
        goto    tens_hd
do_ones
        movfw   hex_lo
        addwf   ones,F                 	;remainder = 1's
	clrf	hex_lo
        return
;
;
;------------------------------------------------------------
;	
;
;	Pulse width attempts to match Goal pulse width 
;
goal_seeker		  		
	movfw 	goal_pw_H		;latest result compared to previous
	subwf 	pw_high,W 		;subtract pw_high - goal_pw_H
	movwf	ACCa_hi			;PW - goal hi result
	skpz				;are they equal ?
	goto 	big_dir			;no
;
	movfw 	goal_pw_L		;yes, compare low
	subwf 	pw_low,W		;subtract pw_low - goal_pw_L
	movwf	ACCa_lo			;PW - goal low result
	skpnz				;if =  Z=1
	return				;goal = PW, leave as is
	skpnc				;
	goto	lil_shrink		;carry set, goal < pw
;			
lil_grow				;goal > PW
	comf	ACCa_lo,F		;invert
	incf	ACCa_lo,W		;
	sublw	.16			;test low result	
	skpc				;
	goto	growing			; > 16 from goal
	movfw 	goal_pw_H		; =< 16
	movwf	pw_high			;
	movfw 	goal_pw_L		;
	movwf	pw_low			;dock PW with Goal
	return
;	
lil_shrink				;PW > goal, as is
	movfw	ACCa_lo			;
	sublw	.16			;test lo result
	skpc				;
	goto	shrinking		; > 16 from goal		
	movfw 	goal_pw_H		; =< 16
	movwf	pw_high			;
	movfw 	goal_pw_L		;
	movwf	pw_low			;dock PW with Goal
	return
;
big_dir
	skpc				; if goal < pw C=1.
	goto	growing			; if pw < goal C=0.
;	
shrinking
	movlw	.16
	subwf	pw_low,F		;shrink by 16
	skpc
	decf	pw_high,F
	return
;
growing
	movlw	.16
	addwf	pw_low,F		;grow by 16
	skpnc
	incf	pw_high,F
	return
;
;------------------------------------------------------------
;               8 x 8 multiply
;               operands in ACCa_lo, ACCb_lo
;               returns hi byte in ACCb_hi
;               returns lo byte in ACCb_lo
multiply_8
        movfw   ACCa_lo
        clrf    ACCb_hi
        clrf    math_count
        bsf     math_count,3
        rrf     ACCb_lo,F
multi_loop8
        skpnc
        addwf   ACCb_hi,F
        rrf     ACCb_hi,F
        rrf     ACCb_lo,F
        decfsz  math_count,F
        goto    multi_loop8
        return
;
;------------------------------------------------------------
;               LCD Routines
;------------------------------------------------------------
;
init_lcd
        bcf     E_LINE                  ;E line low
        bcf     RS_LINE                 ;RS low, control
        call    delay_15mS              ;delay 15 mS
        movlw   0x30                    ;8-bit, 5X7 mode
        movwf   lcd_data
        call    flipbit
        call    pulse
	call	delay_15mS
        movlw   0x20                    ;4-bit, 5x7 mode
        movwf   lcd_data
        call    flipbit
        call    pulse
	call	delay_15mS
        movlw   0x28                    ;4-bit, 5x7 mode
        call    liter_c                 ;send both nybbles
        movlw   0x0c                    ;display on, cursor off
        call    liter_c
        movlw   0x06                    ;increment mode
        call    liter_c
        movlw   0x01                    ;clear display
        call    liter_c
 	call	delay_15mS
        return
;
;------------------------------------------------------------
;		sends control byte to LCD
liter_c					
        movwf   lcd_data                ;store w in lcd_data
	bcf	RS_LINE			;control
        call    delay_125uS
        call    flipbit                 ;hi nibble to PORTB
	call	pulse
        swapf   lcd_data,F              ;swap MS and LS nybbles
        call    flipbit                 ;output what was LS nybble
	call	pulse
	return
;------------------------------------------------------------
;		sends lcd_data to LCD
;	Optional data in W mode available	
liter_d					
;       movwf   lcd_data                ;store w in lcd_data
	bsf	RS_LINE			;data
        call    delay_125uS
        call    flipbit                 ;hi nibble to PORTB
	call	pulse
        swapf   lcd_data,F              ;swap MS and LS nybbles
        call    flipbit                 ;output what was LS nybble
	call	pulse
	return
;
;------------------------------------------------------------
;		clears upper nibble of PORTB then writes
;		upper nibble of lcd_data to PORTB
flipbit
	movlw	0x0f
	andwf	PORTB,F
	movfw	lcd_data
	andlw	0xf0
	iorwf 	PORTB,F
	return
;
;------------------------------------------------------------
pulse
        bsf     E_LINE                  ;pulse E line
        nop                             ;delay
        bcf     E_LINE
        call    delay_125uS            ;delay 125 uS
        return
;
;------------------------------------------------------------
;	Assumes 4 MHz clock
;  	uS delays
;
delay_125uS
	movlw	.40
	movwf	count2			;125 uS delay		
ad_delay_loop
	decfsz	count2,F
	goto	ad_delay_loop
	return	
;
;------------------------------------------------------------
;
delay_15mS
	movlw	.15
	movwf	count2			;15 loops
	clrf	TMR0
	bcf	INTCON,T0IF
delay_15mS_loop
	btfss	INTCON,T0IF		;approx 1 mS per pass
	goto	delay_15mS_loop
	bcf	INTCON,T0IF	
	decfsz	count2,F
	goto	delay_15mS_loop
	return
;
;------------------------------------------------------------
;	with TMR0/4 = 1 mS delay
deb_delay
	clrf	TMR0
	bcf	INTCON,T0IF
deb_loop
	btfss	INTCON,T0IF
	goto	deb_loop
	return
;------------------------------------------------------------
	END
;------------------------------------------------------------	
;Copyright (c) 2012, Mike Berg
;All rights reserved.

;Redistribution and use in source and binary forms, with or without 
;modification, are permitted provided that the following conditions 
;are met:

;Redistributions of source code must retain the above copyright notice, 
;this list of conditions and the following disclaimer. 
;Redistributions in binary form must reproduce the above copyright 
;notice, this list of conditions and the following disclaimer in the 
;documentation and/or other materials provided with the distribution. 
;Neither the name of the <ringolake.com> nor the names of its 
;contributors may be used to endorse or promote products derived from 
;this software without specific prior written permission. 
;THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
;"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
;LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
;FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
;COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
;INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
;BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
;OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
;AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
;OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
;THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
;DAMAGE.
